# Время жизни

Любой объект имеет время жизни. Для упрощения работы с объектами был создан ARC, но даже его наличие не покрывает все нужды. Бывают случаи, когда объект может понадобиться в любое время, но при этом большую часть времени он никому не нужен. В принципе такие объекты являются одним из паттернов программирования и называются "одиночки".

В библиотеке время жизни объявляется у компонентов с помощью метода `lifetime`, который принимает один из возможных вариантов перечисления: `.single`, `.lazySingle`, `.weakSingle`, `.objectGraph`, `.prototype`

## Одиночка (single)
Если объект в программе существует на протяжении всего жизненного цикла программы в единственном экземпляре, то такой объект называется "одиночка". На самом деле в случае с библиотекой объект существует с момента вызова метода `initializeSingletonObjects()` у контейнера и до завершения программы. Чтобы зарегистрировать компонент, который будет иметь единственный экземпляр класса в программе, надо объявить его как `.single`:
```Swift
container.register(Cat.init)
  .lifetime(.single)
```
Если метод  `initializeSingletonObjects()` не будет вызван у контейнера, то любые `.single` компоненты, будут себя вести также как и `.lazySingle`.

## Отложенная одиночка (lazySingle)
Если объект в программе существует на протяжении всего жизненного цикла программы в единственном экземпляре, но при этом создается он при первом обращении, то такой объект называется "отложенная одиночка". Чтобы зарегистрировать компонент, который будет иметь единственный экземпляр класса в программе, но при этом он будет создан при первом обращении, надо объявить его как `.lazySingle`:
```Swift
container.register(Cat.init)
  .lifetime(.lazySingle)
```

## Слабая одиночка (weakSingle)
Если объект в программе существует в единственном экземпляре в любой момент времени пока используется, то такой объект называется "слабая одиночка". То есть два экземпляра объекта не могут существовать одновременно, но если будут потеряны все сильные ссылки на него, то объект удалится, при этом сама библиотека не держит на этот объект сильной ссылки. Чтобы зарегистрировать компонент, который будет иметь единственный экземпляр класса в программе, но при этом он может быть пересоздан, надо объявить его как `.weakSingle`:
```Swift
builder.register(Cat.init)
  .lifetime(.weakSingle)
```

## Единственный в графе (objectGraph)
Объект создается каждый раз новый, но в рамках одного вызова функции `resolve`. То есть, если при создании какого либо объекта, этот объект нужен будет более чем один раз, то это будет один и тотже экземпляр класса. Основное предназначение для чего нужен этот вид времени жизни - циклические зависимости. В случае с циклической зависимость, объект с которого начилось создание, будет присутствовать как миниум два раза: в начале и когда дойдет цикл. От сюда следуют, что при создании цикла хотябы один класс должен иметь время жизни или  `.objectGraph` или любой синглетон. Если быть точнее, то этот "хотябы один класс" должен быть тот класс с которого происходит создание объекта. Не стоит начинать создавать объекты с `.prototype`, несмотря на то, что это явно не запрещено, и отработает, но скорей всего не так как хотелось бы.
```Swift
builder.register(Cat.init)
  .lifetime(.objectGraph)
```


## Всегда новый (prototype)
Самое простое в понимании время жизни - каждый раз при запросе у контейнера объекта, создается новый экземпляр. Во время разрешения зависимостей, также будет создаваться новый экземпляр этого объекта, каждый раз когда это нужно. Объявляется он аналогичным образом:
```Swift
builder.register(Cat.init)
  .lifetime(.prototype)
```

## По умолчанию (default)
Если у компоненты не указано время жизни, то используется `prototype`.


#### [Главная](main.md)
#### [Предыдущая глава "Модификаторы"](modificators.md#Модификаторы)
#### [Следующая глава "Части и Фреймворки"](part_framework.md#Части_и_Фреймворки)
